from ._base import Descriptor
from ._atomic_property import get_mc_gowan_volume

__all__ = ("McGowanVolume",)


class McGowanVolume(Descriptor):
    r"""McGowan volume descriptor.

    References
        * :doi:`10.1007/BF02311772`

    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "McGowan volume"

    @classmethod
    def preset(cls, version):
        yield cls()

    def __str__(self):
        return "VMcGowan"

    def parameters(self):
        return ()

    def calculate(self):
        a = sum(get_mc_gowan_volume(a) for a in self.mol.GetAtoms())
        return a - self.mol.GetNumBonds() * 6.56

    rtype = float



from distutils.version import StrictVersion

from rdkit.Chem import rdMolDescriptors

from ._base import Descriptor
from ._atomic_property import halogen

__all__ = ("AtomCount",)

_desc_conv = {
    "Atom": "all",
    "Bridgehead": "bridgehead",
    "HeavyAtom": "heavy",
    "Hetero": "hetero",
    "Spiro": "spiro",
    "X": "halogen",
}


_version_add_Nhetero = StrictVersion("1.1.0")


class AtomCount(Descriptor):
    r"""atom count descriptor.

    :type type: str
    :param type: type to count.

        * "Atom"
        * "HeavyAtom"
        * "Spiro"
        * "Bridgehead"
        * "Hetero"
        * "X" - all halogen
        * element symbol
    """

    since = "1.0.0"
    __slots__ = ("_type",)

    def description(self):
        return "number of {} atoms".format(_desc_conv.get(self._type, self._type))

    @classmethod
    def preset(cls, version):
        if version >= _version_add_Nhetero:
            t = [
                "Atom",
                "HeavyAtom",
                "Spiro",
                "Bridgehead",
                "Hetero",
                "H",
                "B",
                "C",
                "N",
                "O",
                "S",
                "P",
                "F",
                "Cl",
                "Br",
                "I",
                "X",
            ]
        else:
            t = [
                "Atom",
                "HeavyAtom",
                "Spiro",
                "Bridgehead",
                "H",
                "B",
                "C",
                "N",
                "O",
                "S",
                "P",
                "F",
                "Cl",
                "Br",
                "I",
                "X",
            ]

        return map(cls, t)

    @property
    def explicit_hydrogens(self):
        u"""Require explicit_hydrogens when type is "H" or "Atom"."""
        return self._type in {"H", "Atom"}

    def __str__(self):
        return "n" + self._type

    def parameters(self):
        return (self._type,)

    def __init__(self, type="Atom"):
        self._type = type

    def _calc_X(self):
        X = halogen
        return sum(a.GetAtomicNum() in X for a in self.mol.GetAtoms())

    def _calc(self):
        return sum(a.GetSymbol() == self._type for a in self.mol.GetAtoms())

    def _calc_all(self):
        return self.mol.GetNumAtoms()

    def calculate(self):
        if self._type == "X":
            return self._calc_X()
        elif self._type in ["Atom", "HeavyAtom"]:
            return self._calc_all()
        elif self._type == "Spiro":
            return rdMolDescriptors.CalcNumSpiroAtoms(self.mol)
        elif self._type == "Bridgehead":
            return rdMolDescriptors.CalcNumBridgeheadAtoms(self.mol)
        elif self._type == "Hetero":
            return rdMolDescriptors.CalcNumHeteroatoms(self.mol)
        else:
            return self._calc()

    rtype = int



from rdkit.Chem.rdMolDescriptors import CalcPBF

from ._base import Descriptor

__all__ = ("PBF",)


class PBF(Descriptor):
    r"""PBF descriptor."""

    __slots__ = ()
    since = "1.1.2"
    require_3D = True

    @classmethod
    def preset(cls, version):
        yield cls()

    def description(self):
        return self.__class__.__name__

    def __str__(self):
        return self.__class__.__name__

    def parameters(self):
        return ()

    def calculate(self):
        return CalcPBF(self.get_3D_mol())

    rtype = float



from six import string_types, integer_types
from numpy import product

from ._base import Descriptor
from ._graph_matrix import Valence, DistanceMatrix
from ._atomic_property import GetAtomicNumber, GetElementSymbol

__all__ = ("MolecularDistanceEdge",)


_sp_dict = {1: "primary", 2: "secondary", 3: "tertiary", 4: "quaternary"}


class MolecularDistanceEdge(Descriptor):
    r"""molecular distance edge descriptor.

    :type valence1: int
    :param valence1: valence of first atom

    :type valence2: int
    :param valence2: valence of second atom

    :type element: str or int
    :param element: atomic symbol or atomic number

    :returns: NaN when :math:`N_{\rm MDE} = 0`
    """

    since = "1.0.0"
    __slots__ = ("_valence1", "_valence2", "_atomic_num")
    explicit_hydrogens = False

    def description(self):
        return "molecular distance edge between {a} {e} and {b} {e}".format(
            a=_sp_dict[self._valence1],
            b=_sp_dict[self._valence2],
            e=GetElementSymbol(self._atomic_num),
        )

    @classmethod
    def preset(cls, version):
        return (
            cls(a, b, e)
            for e in [6, 8, 7]
            for a in range(1, 11 - e)
            for b in range(a, 11 - e)
        )

    def __str__(self):
        return "MDE{}-{}{}".format(
            GetElementSymbol(self._atomic_num), self._valence1, self._valence2
        )

    def parameters(self):
        return self._valence1, self._valence2, GetElementSymbol(self._atomic_num)

    def __init__(self, valence1=1, valence2=1, element="C"):
        self._valence1 = min(valence1, valence2)
        self._valence2 = max(valence1, valence2)
        if isinstance(element, integer_types):
            self._atomic_num = element
        elif isinstance(element, string_types):
            self._atomic_num = GetAtomicNumber(element)
        else:
            raise ValueError("element must be atomic number or atomic symbol")

    def dependencies(self):
        return {
            "D": DistanceMatrix(self.explicit_hydrogens),
            "V": Valence(self.explicit_hydrogens),
        }

    def calculate(self, D, V):
        N = self.mol.GetNumAtoms()
        Dv = [
            D[i, j]
            for i in range(N)
            for j in range(i + 1, N)
            if (V[i] == self._valence1 and V[j] == self._valence2)
            or (V[j] == self._valence1 and V[i] == self._valence2)
            if self.mol.GetAtomWithIdx(i).GetAtomicNum()
            == self.mol.GetAtomWithIdx(j).GetAtomicNum()
            == self._atomic_num
        ]
        n = len(Dv)

        with self.rethrow_zerodiv():
            dx = product(Dv) ** (1.0 / (2.0 * n))

        return n / (dx ** 2)

    rtype = float



import inspect
import operator
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from distutils.version import StrictVersion

import six
import numpy as np
from rdkit import Chem

if hasattr(inspect, "getfullargspec"):

    def getargs(func):
        return tuple(inspect.getfullargspec(func).args[1:])


else:

    def getargs(func):
        try:
            return tuple(inspect.getargspec(func).args[1:])
        except TypeError:
            return ()


class MissingValueException(Exception):
    """Internally used exception."""

    __slots__ = ("error",)

    def __init__(self, error):
        self.error = error


class DescriptorMeta(ABCMeta):
    def __new__(cls, classname, bases, dict):
        __init__ = dict.get("__init__")
        if __init__ is None:
            for base in bases:
                __init__ = getattr(base, "__init__", None)
                if __init__ is not None:
                    break

        dict["parameter_names"] = getargs(__init__)
        if "since" in dict:
            dict["since"] = StrictVersion(dict["since"])

        return ABCMeta.__new__(cls, classname, bases, dict)


class Descriptor(six.with_metaclass(DescriptorMeta, object)):
    r"""Abstract base class of descriptors.

    Attributes:
        mol(rdkit.Mol): target molecule

    """

    __slots__ = ("_context",)

    explicit_hydrogens = True
    kekulize = False
    require_connected = False
    require_3D = False

    def __reduce_ex__(self, version):
        return self.__class__, self.parameters()

    def description(self):
        pass

    @classmethod
    def preset(cls, version):
        r"""Generate preset descriptor instances.

        Returns:
            Iterable[Descriptor]: preset descriptors

        """
        return ()

    @abstractmethod
    def parameters(self):
        """[abstractmethod] get __init__ arguments of this descriptor instance.

        this method used in pickling and identifying descriptor instance.

        Returns:
            tuple: tuple of __init__ arguments

        """
        raise NotImplementedError("not implemented Descriptor.parameters method")

    def get_parameter_dict(self):
        return dict(zip(self.parameter_names, self.parameters()))

    def to_json(self):
        """Convert to json serializable dictionary.

        Returns:
            dict: dictionary of descriptor

        """
        d, ps = self._to_json()
        if len(ps) == 0:
            return {"name": d}
        else:
            return {"name": d, "args": ps}

    def _to_json(self):
        d = self.__class__.__name__
        ps = self.get_parameter_dict()

        return d, {k: getattr(v, "as_argument", v) for k, v in ps.items()}

    @abstractmethod
    def calculate(self):
        r"""[abstractmethod] calculate descriptor value.

        Returns:
            rtype

        """
        raise NotImplementedError("not implemented Descriptor.calculate method")

    def dependencies(self):
        r"""Descriptor dependencies.

        Returns:
            dict[str, Descriptor or None] or None

        """
        pass

    @property
    def as_argument(self):
        """Argument representation of descriptor.

        Returns:
            any

        """
        return self

    @staticmethod
    def _pretty(v):
        v = getattr(v, "as_argument", v)
        return repr(v)

    def __repr__(self):
        return "{}.{}({})".format(
            self.__class__.__module__,
            self.__class__.__name__,
            ", ".join(self._pretty(a) for a in self.parameters()),
        )

    def __hash__(self):
        return hash((self.__class__, self.parameters()))

    def __compare_by_reduce(meth):
        def compare(self, other):
            L = self.__class__, self.parameters()
            r = other.__class__, other.parameters()
            return getattr(L, meth)(r)

        return compare

    __eq__ = __compare_by_reduce("__eq__")
    __ne__ = __compare_by_reduce("__ne__")

    __lt__ = __compare_by_reduce("__lt__")
    __gt__ = __compare_by_reduce("__gt__")
    __le__ = __compare_by_reduce("__le__")
    __ge__ = __compare_by_reduce("__ge__")

    rtype = None

    @property
    def mol(self):
        """Get molecule.

        Returns:
            rdkit.Mol

        """
        return self._context.get_mol(self)

    @property
    def config(self):
        return self._context.config

    @property
    def coord(self):
        """Get 3D coordinate.

        Returns:
            numpy.array[3, N]: coordinate matrix

        """
        if not self.require_3D:
            self.fail(AttributeError("use 3D coordinate in 2D descriptor"))

        return self._context.get_coord(self)

    def get_3D_mol(self):
        mol = Chem.Mol(self.mol)
        conf = Chem.Conformer(mol.GetNumAtoms())
        for i, xyz in enumerate(self.coord):
            conf.SetAtomPosition(i, xyz)

        mol.AddConformer(conf)
        return mol

    def fail(self, exception):
        """Raise known exception and return missing value.

        Raises:
            MissingValueException

        """
        raise MissingValueException(exception)

    @contextmanager
    def rethrow_zerodiv(self):
        """[contextmanager] treat zero div as known exception."""
        with np.errstate(divide="raise", invalid="raise"):
            try:
                yield
            except (FloatingPointError, ZeroDivisionError) as e:
                self.fail(ZeroDivisionError(*e.args))

    @contextmanager
    def rethrow_na(self, exception):
        """[contextmanager] treat any exceptions as known exception."""
        try:
            yield
        except exception as e:
            self.fail(e)

    def _unary_common(name, operator):
        def unary(self):
            return UnaryOperatingDescriptor(name.format(self), operator, self)

        return unary

    def _binary_common(name, operator):
        def binary(self, other):
            if not isinstance(other, Descriptor):
                other = ConstDescriptor(other)

            return BinaryOperatingDescriptor(
                name.format(self, other), operator, self, other
            )

        return binary

    __add__ = _binary_common("({}+{})", "+")
    __sub__ = _binary_common("({}-{})", "-")
    __mul__ = _binary_common("({}*{})", "*")
    __truediv__ = _binary_common("({}/{})", "/")
    __floordiv__ = _binary_common("({}//{})", "//")
    __mod__ = _binary_common("({}%{})", "%")
    __pow__ = _binary_common("({}**{})", "**")

    __neg__ = _unary_common("-{}", "-")
    __pos__ = _unary_common("+{}", "+")
    __abs__ = _unary_common("|{}|", "abs")
    __trunc__ = _unary_common("trunc({})", "trunc")

    if six.PY3:
        __ceil__ = _unary_common("ceil({})", "ceil")
        __floor__ = _unary_common("floor({})", "floor")


def is_descriptor_class(desc, include_abstract=False):
    r"""Check calculatable descriptor class or not.

    Returns:
        bool

    """
    return (
        isinstance(desc, type)
        and issubclass(desc, Descriptor)
        and (True if include_abstract else not inspect.isabstract(desc))
    )


class UnaryOperatingDescriptor(Descriptor):
    @classmethod
    def preset(cls, version):
        return cls()

    operators = {
        "+": operator.pos,
        "-": operator.neg,
        "abs": operator.abs,
        "trunc": np.trunc,
        "ceil": np.ceil,  # noqa: S001
        "floor": np.floor,
    }

    def parameters(self):
        return self._name, self._operator, self._value

    def __init__(self, name, operator, value):
        self._name = name
        self._operator = operator
        self._fn = self.operators[operator]
        self._value = value

    def _to_json(self):
        return (
            self.__class__.__name__,
            {
                "name": self._name,
                "operator": self._operator,
                "value": self._value.to_json(),
            },
        )

    def __str__(self):
        return self._name

    def dependencies(self):
        return {"value": self._value}

    def calculate(self, value):
        return self._fn(value)


class ConstDescriptor(Descriptor):
    @classmethod
    def preset(cls, version):
        return cls()

    def parameters(self):
        return (self._value,)

    def __init__(self, value):
        self._value = value

    def __str__(self):
        return str(self._value)

    def calculate(self):
        return self._value


class BinaryOperatingDescriptor(Descriptor):
    @classmethod
    def preset(cls, version):
        return cls()

    operators = {
        "+": operator.add,
        "-": operator.sub,
        "*": operator.mul,  # noqa: S001
        "/": operator.truediv,
        "//": operator.floordiv,
        "%": operator.mod,  # noqa: S001
        "**": operator.pow,
    }

    def _to_json(self):
        return (
            self.__class__.__name__,
            {
                "name": self._name,
                "operator": self._operator,
                "left": self._left.to_json(),  # noqa: S001
                "right": self._right.to_json(),
            },
        )

    def parameters(self):
        return self._name, self._operator, self._left, self._right

    def __init__(self, name, operator, left, right):
        self._name = name
        self._operator = operator
        self._fn = self.operators[operator]
        self._left = left
        self._right = right

    def __str__(self):
        return self._name

    def dependencies(self):
        return {"left": self._left, "right": self._right}

    def calculate(self, left, right):
        return self._fn(left, right)



import numpy as np

from ._base import Descriptor
from ._util import atoms_to_numpy
from ._graph_matrix import AdjacencyMatrix, DistanceMatrix3D

__all__ = ("GravitationalIndex",)


class GravitationalIndex(Descriptor):
    since = "1.0.0"
    __slots__ = "_heavy", "_pair"

    def description(self):
        return "{}{}gravitational index".format(
            "heavy atom " if self._heavy else "", "pair " if self._pair else ""
        )

    @classmethod
    def preset(cls, version):
        return (cls(h, p) for p in [False, True] for h in [True, False])

    require_3D = True

    @property
    def explicit_hydrogens(self):
        return not self._heavy

    def __str__(self):
        return "GRAV{}{}".format("" if self._heavy else "H", "p" if self._pair else "")

    def parameters(self):
        return self._heavy, self._pair

    def __init__(self, heavy=True, pair=False):
        self._heavy = heavy
        self._pair = pair

    def dependencies(self):
        d = {"D": DistanceMatrix3D(self.explicit_hydrogens)}

        if self._pair:
            d["A"] = AdjacencyMatrix(self.explicit_hydrogens)

        return d

    def calculate(self, D, A=1.0):
        w = atoms_to_numpy(lambda a: a.GetMass(), self.mol)

        w = w[:, np.newaxis] * w
        np.fill_diagonal(w, 0)

        D = D.copy()
        np.fill_diagonal(D, 1)

        with self.rethrow_zerodiv():
            return 0.5 * np.sum(w * A / D ** 2)

    rtype = float



"""Mesh generation.

References:
    * http://prideout.net/blog/?p=44
    * http://blog.andreaskahler.com/2009/06/creating-icosphere-mesh-in-code.html

"""

import numpy as np


class SphereMesh(object):
    def __init__(self, level=4):
        t = (1.0 + np.sqrt(5.0)) / 2.0
        self.vertices = np.array(
            [
                (-1, t, 0),
                (1, t, 0),
                (-1, -t, 0),
                (1, -t, 0),
                (0, -1, t),
                (0, 1, t),
                (0, -1, -t),
                (0, 1, -t),
                (t, 0, -1),
                (t, 0, 1),
                (-t, 0, -1),
                (-t, 0, 1),
            ],
            dtype="float",
        )

        self.faces = np.array(
            [
                (0, 11, 5),
                (0, 5, 1),
                (0, 1, 7),
                (0, 7, 10),
                (0, 10, 11),
                (1, 5, 9),
                (5, 11, 4),
                (11, 10, 2),
                (10, 7, 6),
                (7, 1, 8),
                (3, 9, 4),
                (3, 4, 2),
                (3, 2, 6),
                (3, 6, 8),
                (3, 8, 9),
                (4, 9, 5),
                (2, 4, 11),
                (6, 2, 10),
                (8, 6, 7),
                (9, 8, 1),
            ],
            dtype="int",
        )

        self.normalize(0)

        self.level = 1
        self.subdivide(level)

    def normalize(self, begin):
        self.vertices[begin:] /= np.sqrt((self.vertices[begin:] ** 2).sum(axis=1))[
            :, np.newaxis
        ]

    def _subdivide(self):
        self.level += 1

        Nv = len(self.vertices)
        Nf = len(self.faces)

        A = self.faces[:, 0]
        B = self.faces[:, 1]
        C = self.faces[:, 2]

        Av = self.vertices[A]
        Bv = self.vertices[B]
        Cv = self.vertices[C]

        self.vertices = np.r_[self.vertices, Av + Bv, Bv + Cv, Av + Cv]
        self.normalize(Nv)

        AB = np.arange(len(self.faces)) + Nv
        BC = AB + Nf
        AC = AB + 2 * Nf

        self.faces = (
            np.concatenate([A, B, C, AB, AB, AB, AC, AC, AC, BC, BC, BC])
            .reshape(3, -1)
            .T
        )

    def subdivide(self, level):
        for _ in range(level - self.level):
            self._subdivide()

    def __len__(self):
        return len(self.vertices)



from __future__ import division

import numpy as np
from networkx import Graph, floyd_warshall_numpy

from ._base import Descriptor
from ._atomic_property import AtomicProperty, get_properties
from ._matrix_attributes import methods, get_method

__all__ = ("BaryszMatrix",)


class BaryszMatrixBase(Descriptor):
    explicit_hydrogens = False
    __slots__ = ()


class Barysz(BaryszMatrixBase):
    __slots__ = ("_prop",)

    hermitian = True

    def parameters(self):
        return (self._prop,)

    def __init__(self, prop):
        self._prop = prop

    def dependencies(self):
        return {"P": self._prop}

    def calculate(self, P):
        C = self._prop.carbon

        G = Graph()

        G.add_nodes_from(a.GetIdx() for a in self.mol.GetAtoms())

        for bond in self.mol.GetBonds():
            i = bond.GetBeginAtomIdx()
            j = bond.GetEndAtomIdx()

            pi = bond.GetBondTypeAsDouble()

            with self.rethrow_zerodiv():
                w = (C * C) / (P[i] * P[j] * pi)

            G.add_edge(i, j, weight=w)

        sp = floyd_warshall_numpy(G)
        np.fill_diagonal(sp, [1.0 - C / P[a.GetIdx()] for a in self.mol.GetAtoms()])
        return sp


class BaryszMatrix(BaryszMatrixBase):
    r"""barysz matrix descriptor.

    :type prop: :py:class:`str` or :py:class:`function`
    :param prop: :ref:`atomic_properties`

    :type type: str
    :param type: :ref:`matrix_aggregating_methods`

    :returns: NaN when any properties are NaN
    """

    since = "1.0.0"
    __slots__ = ("_prop", "_type")

    def description(self):
        return "{} from Barysz matrix weighted by {}".format(
            self._type.description(), self._prop.get_long()
        )

    @classmethod
    def preset(cls, version):
        return (cls(p, m) for p in get_properties() for m in methods)

    def __str__(self):
        return "{}_Dz{}".format(self._type.__name__, self._prop.as_argument)

    def parameters(self):
        return self._prop, self._type

    def __init__(self, prop="Z", type="SpMax"):
        self._prop = AtomicProperty(self.explicit_hydrogens, prop)
        self._type = get_method(type)

    def dependencies(self):
        return {
            "result": self._type(
                Barysz(self._prop), self.explicit_hydrogens, self.kekulize
            )
        }

    def calculate(self, result):
        return result

    rtype = float



from ._base import Descriptor
from ._graph_matrix import DistanceMatrix

__all__ = ("WienerIndex",)


class WienerIndex(Descriptor):
    r"""Wiener index.

    :type polarity: bool
    :param polarity: use polarity Wiener index
    """

    __slots__ = ("_polarity",)
    since = "1.0.0"
    explicit_hydrogens = False

    def description(self):
        return "Wiener {}index".format("polarity " if self._polarity else "")

    @classmethod
    def preset(cls, version):
        yield cls(False)
        yield cls(True)

    def __str__(self):
        return "WPol" if self._polarity else "WPath"

    def parameters(self):
        return (self._polarity,)

    def __init__(self, polarity=False):
        self._polarity = polarity

    def dependencies(self):
        return {"D": DistanceMatrix(self.explicit_hydrogens)}

    def calculate(self, D):
        if self._polarity:
            return int(0.5 * (D == 3).sum())
        else:
            return int(0.5 * D.sum())

    rtype = int



from rdkit import Chem

from .._util import conformer_to_numpy
from ..error import Missing3DCoordinate


class Context(object):
    __slots__ = "_mols", "_coords", "n_frags", "name", "_stack", "config"

    def __init__(self, mols, coords, n_frags, name, config):
        self._mols = mols
        self._coords = coords
        self.n_frags = n_frags
        self.name = name
        self.config = config

    def __reduce_ex__(self, version):
        return (
            self.__class__,
            (self._mols, self._coords, self.n_frags, self.name, self.config),
        )

    def __str__(self):
        return self.name

    __tf = {True, False}

    @classmethod
    def from_query(cls, mol, require_3D, explicit_hydrogens, kekulizes, id, config):
        if not isinstance(mol, Chem.Mol):
            raise TypeError("{!r} is not rdkit.Chem.Mol instance".format(mol))

        n_frags = len(Chem.GetMolFrags(mol))

        if mol.HasProp("_Name"):
            name = mol.GetProp("_Name")
        else:
            name = Chem.MolToSmiles(Chem.RemoveHs(mol, updateExplicitCount=True))

        mols, coords = {}, {}

        for eh, ke in ((eh, ke) for eh in explicit_hydrogens for ke in kekulizes):
            m = Chem.AddHs(mol) if eh else Chem.RemoveHs(mol, updateExplicitCount=True)

            if ke:
                Chem.Kekulize(m)

            if require_3D:
                try:
                    conf = m.GetConformer(id)
                    if conf.Is3D():
                        coords[eh, ke] = conformer_to_numpy(conf)
                except ValueError:
                    pass

            m.RemoveAllConformers()
            mols[eh, ke] = m

        return cls(mols, coords, n_frags, name, config)

    @classmethod
    def from_calculator(cls, calc, mol, id):
        return cls.from_query(
            mol,
            calc._require_3D,
            calc._explicit_hydrogens,
            calc._kekulizes,
            id,
            calc._config,
        )

    def get_coord(self, desc):
        try:
            return self._coords[desc.explicit_hydrogens, desc.kekulize]
        except KeyError:
            desc.fail(Missing3DCoordinate())

    def get_mol(self, desc):
        return self._mols[desc.explicit_hydrogens, desc.kekulize]

    def reset(self):
        self._stack = []

    def add_stack(self, d):
        self._stack.append(d)

    def get_stack(self):
        return self._stack



from enum import IntEnum
from functools import reduce

from rdkit.Chem import EState

from ._base import Descriptor
from ._util import parse_enum

try:
    import builtins
except ImportError:
    import __builtin__ as builtins


__all__ = ("AtomTypeEState",)


es_types = (
    "sLi",
    "ssBe",
    "ssssBe",
    "ssBH",
    "sssB",
    "ssssB",
    "sCH3",
    "dCH2",
    "ssCH2",
    "tCH",
    "dsCH",
    "aaCH",
    "sssCH",
    "ddC",
    "tsC",
    "dssC",
    "aasC",
    "aaaC",
    "ssssC",
    "sNH3",
    "sNH2",
    "ssNH2",
    "dNH",
    "ssNH",
    "aaNH",
    "tN",
    "sssNH",
    "dsN",
    "aaN",
    "sssN",
    "ddsN",
    "aasN",
    "ssssN",
    "sOH",
    "dO",
    "ssO",
    "aaO",
    "sF",
    "sSiH3",
    "ssSiH2",
    "sssSiH",
    "ssssSi",
    "sPH2",
    "ssPH",
    "sssP",
    "dsssP",
    "sssssP",
    "sSH",
    "dS",
    "ssS",
    "aaS",
    "dssS",
    "ddssS",
    "sCl",
    "sGeH3",
    "ssGeH2",
    "sssGeH",
    "ssssGe",
    "sAsH2",
    "ssAsH",
    "sssAs",
    "sssdAs",
    "sssssAs",
    "sSeH",
    "dSe",
    "ssSe",
    "aaSe",
    "dssSe",
    "ddssSe",
    "sBr",
    "sSnH3",
    "ssSnH2",
    "sssSnH",
    "ssssSn",
    "sI",
    "sPbH3",
    "ssPbH2",
    "sssPbH",
    "ssssPb",
)

es_type_set = set(es_types)


class EStateBase(Descriptor):
    __slots__ = ()
    explicit_hydrogens = False


class EStateCache(EStateBase):
    __slots__ = ()

    def parameters(self):
        return ()

    def calculate(self):
        return EState.TypeAtoms(self.mol), EState.EStateIndices(self.mol)


class AggrType(IntEnum):
    __slots__ = ()

    count = 1
    sum = 2
    max = 3
    min = 4

    @property
    def as_argument(self):
        return self.name

    def description(self):
        d = {self.count: "number", self.sum: "sum", self.max: "max", self.min: "min"}
        return d[self]


aggr_names = (
    (AggrType.count, "N"),
    (AggrType.sum, "S"),
    (AggrType.max, "MAX"),
    (AggrType.min, "MIN"),
)


aggr_name_dict = dict(aggr_names)


class AtomTypeEState(EStateBase):
    r"""atom type e-state descriptor.

    :type type: str
    :param type: one of aggr_types

    :type estate: str
    :param estate: one of es_types

    :returns: NaN when type in ["min", "max"] and :math:`N_{\rm X} = 0`

    References
        * :doi:`10.1021/ci00028a014`

    """

    since = "1.0.0"
    __slots__ = ("_type", "_estate")

    def description(self):
        return "{} of {}".format(self._type.description(), self._estate)

    aggr_types = tuple(a.name for a in AggrType)

    es_types = es_types

    @classmethod
    def preset(cls, version):
        return (cls(a, t) for a in AggrType for t in es_types)

    def __str__(self):
        aggr = aggr_name_dict[self._type]

        return aggr + self._estate

    def parameters(self):
        return self._type, self._estate

    def __init__(self, type="count", estate="sLi"):
        assert estate in es_type_set

        self._type = parse_enum(AggrType, type)
        self._estate = estate

    def dependencies(self):
        return {"E": EStateCache()}

    def calculate(self, E):
        if self._type == AggrType.count:
            return reduce(lambda a, b: a + b, E[0]).count(self._estate)

        indices = map(lambda e: e[1], filter(lambda e: self._estate in e[0], zip(*E)))

        with self.rethrow_na(ValueError):
            return getattr(builtins, self._type.name)(indices)

    @property
    def rtype(self):
        r"""Return type.

        * "count": :py:class:`int`
        * "other": :py:class:`float`
        """
        return int if self._type == AggrType.count else float

    _extra_docs = "aggr_types", "es_types"



from distutils.version import StrictVersion

from ._base import Descriptor
from ._graph_matrix import DistanceMatrix as D
from ._matrix_attributes import SM1, methods, get_method

__all__ = ("DistanceMatrix",)


_version_remove_SM1_D = StrictVersion("1.1.0")


class DistanceMatrix(Descriptor):
    r"""distance matrix descriptor.

    :type type: str
    :param type: :ref:`matrix_aggregating_methods`
    """

    since = "1.0.0"
    __slots__ = ("_type",)
    explicit_hydrogens = False

    def description(self):
        return "{} from distance matrix".format(self._type.description())

    @classmethod
    def preset(cls, version):
        if version >= _version_remove_SM1_D:
            return (cls(m) for m in methods if m != SM1)
        else:
            return map(cls, methods)

    def __str__(self):
        return "{}_D".format(self._type.__name__)

    def parameters(self):
        return (self._type,)

    def __init__(self, type="SpMax"):
        self._type = get_method(type)

    def dependencies(self):
        return {
            "result": self._type(
                D(self.explicit_hydrogens), self.explicit_hydrogens, self.kekulize
            )
        }

    def calculate(self, result):
        return result

    rtype = float



import inspect

from mordred import descriptors, get_descriptors_in_module
from nose.tools import ok_
from mordred._base.descriptor import is_descriptor_class


def test_attributes():
    for mdl in descriptors.all:
        for desc in get_descriptors_in_module(mdl, submodule=False):
            for cls in desc.__mro__:
                if cls == object:
                    continue

                if is_descriptor_class(desc, include_abstract=True):
                    yield (
                        ok_,
                        "__slots__" in cls.__dict__,
                        "{}({}) class don't have __slots__".format(
                            cls.__name__, inspect.getfile(desc)
                        ),
                    )

            yield (
                ok_,
                "since" in desc.__dict__,
                "{}({}) class don't have since".format(
                    desc.__name__, inspect.getfile(desc)
                ),
            )



from collections import Iterator

from mordred import (
    Calculator,
    descriptors,
    get_descriptors_in_module,
    get_descriptors_from_module,
)
from nose.tools import ok_


def test_descriptor_order():
    calc = Calculator(descriptors)
    it = iter(calc.descriptors)
    before = next(it).__module__
    for current in it:
        current = current.__module__
        assert before <= current, "{!r} > {!r}".format(before, current)

        before = current


def test_get_descriptors_in_module():
    old = get_descriptors_from_module(descriptors, True)
    new = get_descriptors_in_module(descriptors, True)

    yield isinstance, new, Iterator
    yield ok_, len(list(new)) > 100

    for a, b in zip(old, new):
        assert a == b



from rdkit import Chem
from numpy.testing import assert_almost_equal

from mordred import _atomic_property

_data = {
    ">C<": "CC(C)(C)C",
    ">CH-": "CC(C)C",
    "-CH2-": "CCC",  # noqa: S001
    "=C<": "C=C(C)C",
    "-CH3": "CC",  # noqa: S001
    "=CH-": "CC=C",
    ">N-": "CN(C)C",
    "#C-": "C#CC",  # noqa: S001
    "-NH-": "CNC",
    "=CH2": "C=C",
    "=N-": "C=NC",
    "-O-": "COC",  # noqa: S001
    "#CH": "C#C",  # noqa: S001
    "-NH2": "CN",
    "=NH": "C=N",
    "#N": "C#N",  # noqa: S001
    "-OH": "CO",
    "=O": "C=O",
    "-F": "CF",  # noqa: S001
    "-SH": "CS",
    "-S-": "CSC",  # noqa: S001
    "=S": "C=S",
    "-Cl": "CCl",  # noqa: S001
    "-Br": "CBr",  # noqa: S001
    "-I": "CI",
}

explicitHs = {
    key: Chem.AddHs(Chem.MolFromSmiles(smi)).GetAtomWithIdx(1)
    for key, smi in _data.items()
}

implicitHs = {
    key: Chem.MolFromSmiles(smi).GetAtomWithIdx(1) for key, smi in _data.items()
}


def make_cases(results, fn, decimal=7):
    for key, result in results.items():
        yield assert_almost_equal, result, fn(explicitHs[key]), decimal, key
        yield assert_almost_equal, result, fn(implicitHs[key]), decimal, key


def test_sigma():
    results = {
        ">C<": 4,
        ">CH-": 3,
        "-CH2-": 2,  # noqa: S001
        "=C<": 3,
        "-CH3": 1,  # noqa: S001
        "=CH-": 2,
        ">N-": 3,
        "#C-": 2,  # noqa: S001
        "-NH-": 2,
        "=CH2": 1,
        "=N-": 2,
        "-O-": 2,  # noqa: S001
        "#CH": 1,  # noqa: S001
        "-NH2": 1,
        "=NH": 1,
        "#N": 1,  # noqa: S001
        "-OH": 1,
        "=O": 1,
        "-F": 1,  # noqa: S001
        "-SH": 1,
        "-S-": 2,  # noqa: S001
        "=S": 1,
        "-Cl": 1,  # noqa: S001
        "-Br": 1,  # noqa: S001
        "-I": 1,
    }
    for test in make_cases(results, _atomic_property.get_sigma_electrons):
        yield test


def test_valence_sigma():
    results = {
        ">C<": 4,
        ">CH-": 3,
        "-CH2-": 2,  # noqa: S001
        "=C<": 4,
        "-CH3": 1,  # noqa: S001
        "=CH-": 3,
        ">N-": 5,
        "#C-": 4,  # noqa: S001
        "-NH-": 4,
        "=CH2": 2,
        "=N-": 5,
        "-O-": 6,  # noqa: S001
        "#CH": 3,  # noqa: S001
        "-NH2": 3,
        "=NH": 4,
        "#N": 5,  # noqa: S001
        "-OH": 5,
        "=O": 6,
        "-F": 7,  # noqa: S001
        "-S-": 0.67,
        "-Cl": 0.78,  # noqa: S001
        "-Br": 0.26,  # noqa: S001
        "-I": 0.16,
    }
    for test in make_cases(results, _atomic_property.get_valence_electrons, 2):
        yield test


def test_IntrinsicState():
    results = {
        ">C<": 1.25,
        ">CH-": 1.3333,
        "-CH2-": 1.5,  # noqa: S001
        "=C<": 1.6666,
        "-CH3": 2.0,  # noqa: S001
        "=CH-": 2.0,
        ">N-": 2.0,
        "#C-": 2.5,  # noqa: S001
        "-NH-": 2.5,
        "=CH2": 3.0,
        "=N-": 3.0,
        "-O-": 3.5,  # noqa: S001
        "#CH": 4.0,  # noqa: S001
        "-NH2": 4.0,
        "=NH": 5.0,
        "#N": 6.0,  # noqa: S001
        "-OH": 6.0,
        "=O": 7.0,
        "-F": 8.0,  # noqa: S001
    }

    for test in make_cases(results, _atomic_property.get_intrinsic_state, 3):
        yield test



import os

import nose


def main():
    base = os.path.dirname(os.path.dirname(__file__))
    hidden = [
        os.path.join(base, n)
        for n in os.listdir(base)
        if n[:1] == "_" and os.path.splitext(n)[1] == ".py"
    ]

    tests = [base, os.path.join(base, "_base")] + hidden

    os.environ["NOSE_WITH_DOCTEST"] = "1"

    nose.main(defaultTest=",".join(tests))


if __name__ == "__main__":
    main()



from rdkit import Chem

from mordred import Calculator, Descriptor
from nose.tools import eq_
from mordred.error import Error


class RaiseDescriptor(Descriptor):
    def parameters(self):
        return ()

    def __init__(self, e, critical):
        self.e = e
        self.critical = critical

    def calculate(self):
        raise self.e


mol = Chem.MolFromSmiles("c1ccccc1")


def test_catch_non_critical_error():
    calc = Calculator(RaiseDescriptor(ValueError("test exception"), False))

    result = calc(mol)[0]
    assert isinstance(result, Error)
    eq_(result.error.args, ("test exception",))



r"""charged partial surface area descriptor.

References
    * :doi:`10.1021/ac00220a013`

"""
import numpy as np

from ._base import Descriptor
from ._util import atoms_to_numpy
from .surface_area import SurfaceArea
from ._atomic_property import AtomicProperty, vdw_radii

__all__ = (
    "PNSA",
    "PPSA",
    "DPSA",
    "FNSA",
    "FPSA",
    "WNSA",
    "WPSA",
    "RNCG",
    "RPCG",
    "RNCS",
    "RPCS",
    "TASA",
    "TPSA",
    "RASA",
    "RPSA",
)


class CPSABase(Descriptor):
    __slots__ = ()
    require_3D = True

    @classmethod
    def preset(cls, version):
        yield cls()

    def parameters(self):
        return ()

    def __str__(self):
        return self.__class__.__name__

    rtype = float


class VersionCPSABase(CPSABase):
    __slots__ = ("_version",)

    @classmethod
    def preset(cls, version):
        return map(cls, cls.versions)

    def parameters(self):
        return (self._version,)

    versions = [1, 2, 3, 4, 5]

    def __init__(self, version=1):
        self._version = version

    def __str__(self):
        return "{}{}".format(self.__class__.__name__, self._version)


class AtomicSurfaceArea(CPSABase):
    __slots__ = "_solvent_radius", "_level"

    def parameters(self):
        return self._solvent_radius, self._level

    def __init__(self, solvent_radius=1.4, level=5):
        self._solvent_radius = solvent_radius
        self._level = level

    def calculate(self):
        rs = atoms_to_numpy(
            lambda a: vdw_radii[a.GetAtomicNum()] + self._solvent_radius, self.mol
        )

        with self.rethrow_zerodiv():
            sa = SurfaceArea(rs, self.coord, self._level)
        return np.array(sa.surface_area())

    rtype = None


class TotalSurfaceArea(CPSABase):
    __slots__ = ()

    def dependencies(self):
        return {"ASA": AtomicSurfaceArea()}

    def calculate(self, ASA):
        return np.sum(ASA)


class AtomicCharge(CPSABase):
    __slots__ = ()
    require_3D = False

    def dependencies(self):
        return {"charges": AtomicProperty(self.explicit_hydrogens, "c")}

    def calculate(self, charges):
        return charges

    rtype = None


class PNSA(VersionCPSABase):
    r"""partial negative surface area descriptor.

    .. math::

        {\rm PNSA}_1 = \sum_{a-} {\rm SA}_a^-

    where :math:`\sum_{a-}` means sum over negative charged atoms,
    :math:`{\rm SA}_a^-` is atomic partial surface area.

    :type version: int
    :param version: one of :py:attr:`versions`
    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "partial negative surface area (version {})".format(self._version)

    def dependencies(self):
        return {"SA": AtomicSurfaceArea(), "charges": AtomicCharge()}

    @staticmethod
    def _mask(charges):
        return charges < 0.0

    def calculate(self, SA, charges):
        mask = self._mask(charges)

        if self._version == 1:
            f = 1.0
        elif self._version == 2:
            f = np.sum(charges[mask])
        elif self._version == 3:
            f = charges[mask]
        elif self._version == 4:
            f = np.sum(charges[mask]) / self.mol.GetNumAtoms()
        elif self._version == 5:
            with self.rethrow_zerodiv():
                f = np.sum(charges[mask]) / np.sum(mask)

        return np.sum(f * SA[mask])


class PPSA(PNSA):
    r"""partial positive surface area descriptor.

    :type version: int
    :param version: one of :py:attr:`versions`
    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "partial positive surface area (version {})".format(self._version)

    @staticmethod
    def _mask(charges):
        return charges > 0.0


class DPSA(VersionCPSABase):
    r"""difference in charged partial surface area descriptor.

    :type version: int
    :param version: one of :py:attr:`versions`
    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "difference in charged partial surface area (version {})".format(
            self._version
        )

    def dependencies(self):
        return {"PNSA": PNSA(self._version), "PPSA": PPSA(self._version)}

    def calculate(self, PPSA, PNSA):
        return PPSA - PNSA


class FNSA(VersionCPSABase):
    r"""fractional charged partial negative surface area descriptor.

    :type version: int
    :param version: one of :py:attr:`versions`
    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "fractional charged partial negative surface area (version {})".format(
            self._version
        )  # noqa: E501

    def _SA(self):
        return PNSA(self._version)

    def dependencies(self):
        return {"ASA": AtomicSurfaceArea(), "SA": self._SA()}

    def calculate(self, SA, ASA):
        return SA / np.sum(ASA)


class FPSA(FNSA):
    r"""fractional charged partial positive surface area descriptor.

    :type version: int
    :param version: one of :py:attr:`versions`
    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "fractional charged partial positive surface area (version {})".format(
            self._version
        )  # noqa: E501

    def _SA(self):
        return PPSA(self._version)


class WNSA(FNSA):
    r"""surface weighted charged partial negative surface area descriptor.

    :type version: int
    :param version: one of :py:attr:`versions`
    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "surface weighted charged partial negative surface area (version {})".format(
            self._version
        )  # noqa: E501

    def calculate(self, SA, ASA):
        return SA * np.sum(ASA) / 1000.0


class WPSA(FPSA):
    r"""surface weighted charged partial positive surface area descriptor.

    :type version: int
    :param version: one of :py:attr:`versions`
    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "surface weighted charged partial positive surface area (version {})".format(
            self._version
        )  # noqa: E501

    def calculate(self, SA, ASA):
        return SA * np.sum(ASA) / 1000.0


class RNCG(CPSABase):
    r"""relative negative charge descriptor."""

    since = "1.0.0"
    __slots__ = ()
    require_3D = False

    def description(self):
        return "relative negative charge"

    @staticmethod
    def _mask(charges):
        return charges < 0.0

    def dependencies(self):
        return {"charges": AtomicCharge()}

    def calculate(self, charges):
        charges = charges[self._mask(charges)]
        if len(charges) == 0:
            return 0.0

        Qmax = charges[np.argmax(np.abs(charges))]

        return Qmax / np.sum(charges)


class RPCG(RNCG):
    r"""relative positive charge descriptor."""

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "relative positive charge"

    @staticmethod
    def _mask(charges):
        return charges > 0.0


class RNCS(CPSABase):
    r"""relative negative charge surface area descriptor."""

    since = "1.0.0"
    __slots__ = ()
    _RCG = RNCG()

    def description(self):
        return "relative negative charge surface area"

    def dependencies(self):
        return {"RCG": self._RCG, "SA": AtomicSurfaceArea(), "charges": AtomicCharge()}

    @staticmethod
    def _mask(charges):
        return charges < 0

    def calculate(self, RCG, SA, charges):
        mask = self._mask(charges)
        charges = charges[mask]
        if len(charges) == 0:
            return 0.0

        SAmax = SA[mask][np.argmax(np.abs(charges))]

        return SAmax / RCG

    rtype = float


class RPCS(RNCS):
    r"""relative positive charge surface area descriptor."""

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "relative positive charge surface area"

    @staticmethod
    def _mask(charges):
        return charges > 0

    _RCG = RPCG()


class TASA(CPSABase):
    r"""total hydrophobic surface area descriptor."""

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "total hydrophobic surface area"

    @staticmethod
    def _mask(charges):
        return np.abs(charges) < 0.2

    def dependencies(self):
        return {"SA": AtomicSurfaceArea(), "charges": AtomicCharge()}

    def calculate(self, SA, charges):
        return np.sum(SA[self._mask(charges)])


class TPSA(TASA):
    r"""total polar surface area descriptor."""

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "total polar surface area"

    @staticmethod
    def _mask(charges):
        return np.abs(charges) >= 0.2


class RASA(CPSABase):
    r"""relative hydrophobic surface area descriptor."""

    since = "1.0.0"
    __slots__ = ()
    _TxSA = TASA()

    def description(self):
        return "relative hydrophobic surface area"

    def dependencies(self):
        return {"SASA": AtomicSurfaceArea(), "TxSA": self._TxSA}

    def calculate(self, TxSA, SASA):
        return TxSA / np.sum(SASA)


class RPSA(RASA):
    r"""relative polar surface area descriptor."""

    def description(self):
        return "relative polar surface area"

    since = "1.0.0"
    __slots__ = ()
    _TxSA = TPSA()



"""Error objects."""

from abc import ABCMeta, abstractproperty

import six
import numpy as np


class MissingValueBase(six.with_metaclass(ABCMeta, object)):
    """Base class of missing values.

    Args:
        error (Exception): error object
        stack (callstack)

    """

    __slots__ = "error", "stack"

    def __reduce_ex__(self, version):
        return self.__class__, (self.error, self.stack)

    def __init__(self, error, stack):
        self.error = error
        self.stack = stack

    def __float__(self):
        return np.nan

    def __add__(self, other):
        return np.nan

    def __sub__(self, other):
        return np.nan

    def __str__(self):
        return "{} ({})".format(self.error, "/".join(str(d) for d in self.stack))

    @abstractproperty
    def header(self):
        """Header of warning message.

        Returns:
            str

        """
        raise NotImplementedError("require header")


class Missing(MissingValueBase):
    """known errored value."""

    __slots__ = ()
    header = "Missing"


class Error(MissingValueBase):
    """unknown errored value."""

    __slots__ = ()
    header = "ERROR"


class MordredException(Exception):
    __slots__ = ()


class MultipleFragments(MordredException):
    """multiple fragments detected on require_connected Descriptor."""

    __slots__ = ()

    def __str__(self):
        return "multiple fragments"


class Missing3DCoordinate(MordredException):
    """missing 3D coordinate on require_3D Descriptor."""

    __slots__ = ()

    def __str__(self):
        return "missing 3D coordinate"


class DuplicatedDescriptorName(MordredException):
    """duplicated string replisantation of descriptor."""

    __slots__ = ()

    def __init__(self, a, b):
        self.a = a
        self.b = b

    def __str__(self):
        return "duplicated descriptor name: {!r} and {!r}".format(self.a, self.b)


class Timeout(MordredException):
    """calculation timed out."""

    __slots__ = ()

    def __str__(self):
        return "timed out"



"""Mordred base package."""

from .util import is_missing
from ..error import MissingValueBase
from .result import Result
from .parallel import parallel
from .calculator import (
    Calculator,
    get_descriptors_in_module,
    get_descriptors_from_module,
)
from .descriptor import (
    Descriptor,
    ConstDescriptor,
    UnaryOperatingDescriptor,
    BinaryOperatingDescriptor,
)

__all__ = (
    "Descriptor",
    "Calculator",
    "get_descriptors_from_module",
    "get_descriptors_in_module",
    "is_missing",
    "Result",
)


def _Descriptor__call__(self, mol, id=-1):
    r"""Calculate single descriptor value.

    :type id: int
    :param id: conformer id

    :returns: descriptor result
    :rtype: scalar
    """
    v = Calculator(self)(mol, id)[0]
    if isinstance(v, MissingValueBase):
        raise v.error

    return v


def _from_json(obj, descs):
    name = obj.get("name")
    args = obj.get("args") or {}
    if name is None:
        raise ValueError("invalid json: {}".format(obj))

    if name == UnaryOperatingDescriptor.__name__:
        return UnaryOperatingDescriptor(
            args["name"], args["operator"], _from_json(args["value"])
        )

    elif name == BinaryOperatingDescriptor.__name__:
        return BinaryOperatingDescriptor(
            args["name"],
            args["operator"],
            _from_json(args["left"]),
            _from_json(args["right"]),
        )

    cls = descs.get(name)
    if cls is None:
        raise ValueError("unknown class: {}".format(name))

    instance = cls(**(obj.get("args") or {}))
    return instance


@classmethod
def _Descriptor_from_json(self, obj):
    """Create Descriptor instance from json dict.

    Parameters:
        obj(dict): descriptor dict

    Returns:
        Descriptor: descriptor

    """
    descs = getattr(self, "_all_descriptors", None)

    if descs is None:
        from mordred import descriptors

        descs = {cls.__name__: cls for cls in get_descriptors_in_module(descriptors)}
        descs[ConstDescriptor.__name__] = ConstDescriptor
        self._all_descriptors = descs

    return _from_json(obj, descs)


Descriptor.__call__ = _Descriptor__call__
Descriptor.from_json = _Descriptor_from_json
Calculator._parallel = parallel



from rdkit.Chem.Descriptors import MolWt, ExactMolWt

from ._base import Descriptor

__all__ = ("Weight",)


class Weight(Descriptor):
    r"""molecular weight descriptor.

    :type averaged: bool
    :param averaged: averaged by number of atom
    """

    def description(self):
        return "{}{}molecular weight".format(
            "averaged " if self._averaged else "", "exact " if self._exact else ""
        )

    since = "1.0.0"
    __slots__ = ("_averaged", "_exact")
    explicit_hydrogens = True

    @classmethod
    def preset(cls, version):
        yield cls(True, False)
        yield cls(True, True)

    def __str__(self):
        return "{}{}MW".format(
            "A" if self._averaged else "", "" if self._exact else "a"
        )

    def parameters(self):
        return self._exact, self._averaged

    def __init__(self, exact=True, averaged=False):
        self._averaged = averaged
        self._exact = exact

    def calculate(self):
        w = ExactMolWt(self.mol) if self._exact else MolWt(self.mol)
        if self._averaged:
            w /= self.mol.GetNumAtoms()

        return w

    rtype = float



from enum import IntEnum
from itertools import chain

from rdkit import Chem

from ._base import Descriptor
from ._util import parse_enum

__all__ = ("BondCount",)


class BondType(IntEnum):
    __slots__ = ()

    any = 1
    heavy = 2

    single = 3
    double = 4
    triple = 5

    aromatic = 6
    multiple = 7

    @property
    def as_argument(self):
        return self.name


bond_types = (
    (BondType.any, ("", "all bonds", lambda _: True)),
    (BondType.heavy, ("O", "bonds connecting to heavy atom", lambda _: True)),
    (
        BondType.single,
        ("S", "single bonds", lambda b: b.GetBondType() == Chem.BondType.SINGLE),
    ),
    (
        BondType.double,
        ("D", "double bonds", lambda b: b.GetBondType() == Chem.BondType.DOUBLE),
    ),
    (
        BondType.triple,
        ("T", "triple bonds", lambda b: b.GetBondType() == Chem.BondType.TRIPLE),
    ),
    (
        BondType.aromatic,
        (
            "A",
            "aromatic bonds",
            lambda b: b.GetIsAromatic() or b.GetBondType() == Chem.BondType.AROMATIC,
        ),
    ),
    (
        BondType.multiple,
        (
            "M",
            "multiple bonds",
            lambda b: b.GetIsAromatic() or b.GetBondType() != Chem.BondType.SINGLE,
        ),
    ),
)

bond_type_dict = dict(bond_types)


class BondCount(Descriptor):
    r"""bond count descriptor.

    :type type: str
    :param type: one of bond_types

    :type kekulize: bool
    :param kekulize: use kekulized structure
    """

    since = "1.0.0"
    __slots__ = ("_type", "_bond_name", "_bond_desc", "_check_bond", "kekulize")

    def description(self):
        return "number of {} in {}kekulized structure".format(
            self._bond_desc, "" if self.kekulize else "non-"
        )

    bond_types = tuple(b.name for b in BondType)

    @classmethod
    def preset(cls, version):
        return chain(
            map(lambda t: cls(t, False), BondType),
            map(lambda t: cls(t, True), [BondType.single, BondType.double]),
        )

    def __str__(self):
        K = "K" if self.kekulize else ""
        return "nBonds{}{}".format(K, self._bond_name)

    @property
    def explicit_hydrogens(self):
        return self._type in (BondType.any, BondType.single)

    def parameters(self):
        return self._type, self.kekulize

    def __init__(self, type="any", kekulize=False):
        self._type = parse_enum(BondType, type)
        self._bond_name, self._bond_desc, self._check_bond = bond_type_dict[self._type]
        self.kekulize = kekulize

    def calculate(self):
        return sum(1 for b in self.mol.GetBonds() if self._check_bond(b))

    rtype = int

    _extra_docs = ("bond_types",)



from __future__ import print_function

import argparse

from rdkit import Chem

from . import SurfaceArea
from .._util import PathType, module_prog


def main():
    parser = argparse.ArgumentParser(
        prog=module_prog(__package__), formatter_class=argparse.MetavarTypeHelpFormatter
    )
    parser.add_argument("sdf", type=PathType, help="input sd file", metavar="SDF")
    parser.add_argument(
        "-s",
        "--solvent-radius",
        type=float,
        default=1.4,
        help="solvent radius (default: %(default)s)",
    )
    parser.add_argument(
        "-l",
        "--mesh-level",
        type=int,
        default=5,
        help="mesh level (default: %(default)s)",
    )
    result = parser.parse_args()
    main_process(
        sdf=result.sdf,
        solvent_radius=result.solvent_radius,
        mesh_level=result.mesh_level,
    )


def main_process(sdf, solvent_radius, mesh_level):
    for i, mol in enumerate(Chem.SDMolSupplier(sdf, removeHs=False)):
        name = mol.GetProp("_Name") if mol.HasProp("_Name") else str(i)
        sa = SurfaceArea.from_mol(mol, solvent_radius=solvent_radius, level=mesh_level)

        print(name)  # noqa: T003

        for a, s in zip(mol.GetAtoms(), sa.surface_area()):
            print(  # noqa: T001
                "{:4d} {:2s} {:8.3f}".format(a.GetIdx(), a.GetSymbol(), s)
            )  # noqa: T003


if __name__ == "__main__":
    main()



from rdkit import Chem
from numpy.testing import assert_almost_equal

from mordred.SLogP import SMR, SLogP


def test_WildmanCrippen1():
    mol = Chem.MolFromSmiles("Oc1ccccc1OC")
    yield assert_almost_equal, SLogP()(mol), 1.4, 2
    yield assert_almost_equal, SMR()(mol), 34.66, 2


def test_WildmanCrippen2():
    mol = Chem.MolFromSmiles("c1ccccc1c2ccccn2")
    yield assert_almost_equal, SLogP()(mol), 2.75, 2



from ._base import Descriptor
from .SLogP import SMR, SLogP
from .Weight import Weight
from .HydrogenBond import HBondDonor, HBondAcceptor

__all__ = ("Lipinski", "GhoseFilter")


class LipinskiLike(Descriptor):
    __slots__ = ()

    @classmethod
    def preset(cls, version):
        yield cls()

    def parameters(self):
        return ()

    def __str__(self):
        return self.__class__.__name__

    rtype = bool


class Lipinski(LipinskiLike):
    r"""Lipinski rule of 5 descriptor.

    LogP: SLogP
    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "Lipinski rule of five"

    def dependencies(self):
        return {
            "HBAcc": HBondAcceptor(),
            "HBDon": HBondDonor(),
            "LogP": SLogP(),
            "MW": Weight(),
        }

    def calculate(self, LogP, MW, HBDon, HBAcc):
        return HBDon <= 5 and HBAcc <= 10 and MW <= 500 and LogP <= 5


class GhoseFilter(LipinskiLike):
    r"""Ghose filter descriptor.

    LogP, MR: SLogP, SMR
    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "Ghose filter"

    def dependencies(self):
        return {"LogP": SLogP(), "MR": SMR(), "MW": Weight()}

    def calculate(self, MW, LogP, MR):
        return (
            (160 <= MW <= 480)
            and (20 <= self.mol.GetNumAtoms() <= 70)
            and (-0.4 <= LogP <= 5.6)
            and (40 <= MR <= 130)
        )



"""Acid Base descriptor.

References:
    * http://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/AcidicGroupCountDescriptor.html
    * http://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/BasicGroupCountDescriptor.html

"""  # noqa: E501

from abc import abstractproperty

from rdkit import Chem

from ._base import Descriptor

__all__ = ("AcidicGroupCount", "BasicGroupCount")


class SmartsCountBase(Descriptor):
    __slots__ = ("_mol",)

    @classmethod
    def preset(cls, version):
        yield cls()

    def _create_smarts(self):
        s = ",".join("$(" + s + ")" for s in self.SMARTS)
        self._mol = Chem.MolFromSmarts("[" + s + "]")
        return self._mol

    @abstractproperty
    def SMARTS(self):
        pass

    def __str__(self):
        return self._name

    def parameters(self):
        return ()

    def calculate(self):
        pat = getattr(self, "_mol", None) or self._create_smarts()
        return len(self.mol.GetSubstructMatches(pat))

    rtype = int

    _extra_docs = ("SMARTS",)


class AcidicGroupCount(SmartsCountBase):
    r"""acidic group count descriptor."""

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "acidic group count"

    _name = "nAcid"

    SMARTS = (
        "[O;H1]-[C,S,P]=O",
        "[*;-;!$(*~[*;+])]",
        "[NH](S(=O)=O)C(F)(F)F",
        "n1nnnc1",
    )


class BasicGroupCount(SmartsCountBase):
    r"""basic group count descriptor."""

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "basic group count"

    _name = "nBase"

    SMARTS = (
        "[NH2]-[CX4]",
        "[NH](-[CX4])-[CX4]",
        "N(-[CX4])(-[CX4])-[CX4]",
        "[*;+;!$(*~[*;-])]",
        "N=C-N",
        "N-C=N",
    )



from __future__ import print_function

import os
import sys
import logging
import argparse
from importlib import import_module
from multiprocessing import freeze_support

from rdkit import Chem

from . import Calculator, __version__, descriptors
from ._base import get_descriptors_in_module
from ._util import PathType, module_prog
from .error import Missing, MissingValueBase


def smiles_parser(path):
    with open(path) as file:
        for line in file:
            line = line.strip().split()
            if len(line) == 1:
                smi = line[0]
                name = smi
            else:
                smi = line[0]
                name = " ".join(line[1:])

            mol = Chem.MolFromSmiles(smi)

            if mol is None:
                logging.warning("smiles read failure: %s", name)
                continue

            mol.SetProp("_Name", name)
            yield mol


def sdf_parser(path):
    base = os.path.splitext(os.path.basename(path))[0]

    for i, mol in enumerate(Chem.SDMolSupplier(path, removeHs=False)):
        if mol is None:
            logging.warning("mol read failure: %s.%s", base, i)
            continue

        if mol.GetProp("_Name") == "":
            mol.SetProp("_Name", "{}.{}".format(base, i))

        yield mol


def auto_parser(path):
    ext = os.path.splitext(path)[1]
    if ext == ".smi":
        r = smiles_parser(path)
    elif ext in [".mol", ".sdf"]:
        r = sdf_parser(path)
    else:
        logging.warning("cannot detect file format: %s", path)
        r = ()

    for m in r:
        yield m


class ParserAction(argparse.Action):
    def __init__(self, option_strings, dest, **kwargs):
        super(ParserAction, self).__init__(option_strings, dest, **kwargs)
        self.default = self.to_parser("auto")
        self.choices = ["auto", "sdf", "mol", "smi"]

    def to_parser(self, value):
        if value == "auto":
            return auto_parser
        elif value == "smi":
            return smiles_parser
        elif value in ["sdf", "mol"]:
            return sdf_parser

        raise ValueError("invalid parser: {}".format(value))

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, self.to_parser(values))


def make_parser():
    parser = argparse.ArgumentParser(
        prog=module_prog(__package__),
        epilog="descriptors: {}".format(" ".join(descriptors.__all__)),
    )
    parser.add_argument(
        "--version",
        action="version",
        help="input molecular file",
        version="{}-{}".format(__package__, __version__),
    )
    parser.add_argument("input", type=PathType, nargs="+", metavar="INPUT")
    parser.add_argument(
        "-t", "--type", action=ParserAction, help="input filetype (default: auto)"
    )
    parser.add_argument(
        "-o",
        "--output",
        default="-",
        type=argparse.FileType("w"),
        help="output file path (default: stdout)",
    )
    parser.add_argument(
        "-p",
        "--processes",
        default=None,
        type=int,
        help="number of processes (default: number of logical processors)",
    )
    parser.add_argument("-q", "--quiet", action="store_true", help="hide progress bar")
    parser.add_argument("-s", "--stream", action="store_true", help="stream read")
    parser.add_argument(
        "-d",
        "--descriptor",
        default=[],
        choices=descriptors.__all__,
        action="append",
        help="descriptors to calculate (default: all)",
        metavar="DESC",
    )
    parser.add_argument(
        "-3",
        "--3D",
        action="store_true",
        dest="with3D",
        help="use 3D descriptors (require sdf or mol file)",
    )
    parser.add_argument(
        "-v", "--verbosity", action="count", default=0, help="verbosity"
    )

    return parser


def main_process(
    input, parser, output, nproc, quiet, stream, descriptor, with3D, verbosity
):
    mols = (m for i in input for m in parser(i))

    if output.isatty():
        quiet = True

    if stream:
        N = None
    else:
        mols = list(mols)
        N = len(mols)

    # Descriptors
    calc = Calculator()

    if verbosity >= 2:
        calc._debug = True

    if len(descriptor) == 0:
        calc.register(descriptors, ignore_3D=not with3D)
    else:
        calc.register(
            (
                d
                for m in descriptor
                for d in get_descriptors_in_module(
                    import_module("." + m, __package__), False
                )
            ),
            ignore_3D=not with3D,
        )

    with output:
        write_row(output, ["name"] + [str(d) for d in calc.descriptors])

        def warning(name, v, err_set):
            if not isinstance(v, MissingValueBase):
                return

            if verbosity == 0 and isinstance(v, Missing):
                return

            red = v.error.__class__, v.error.args
            if red in err_set:
                return

            calc.echo("[{}] {}: {}".format(v.header, name, v), file=sys.stderr)
            err_set.add(red)

        def pretty(name, v, err_set):
            warning(name, v, err_set)

            if isinstance(v, MissingValueBase):
                return ""

            if isinstance(v, bool):
                return int(v)

            return str(v)

        for result in calc.map(mols, nproc=nproc, nmols=N, quiet=quiet):
            err_set = set()

            if result.mol.HasProp("_Name"):
                name = result.mol.GetProp("_Name")
            else:
                name = Chem.MolToSmiles(result.mol)

            write_row(output, [name] + [pretty(name, v, err_set) for v in result])


def write_row(file, data):
    file.write(
        ",".join(
            str(v).replace('"', '""').replace("\n", "").replace("\r", "") for v in data
        )
    )
    file.write("\n")


def main(args=None):
    parser = make_parser()
    p = parser.parse_args(args)
    return main_process(
        input=p.input,
        parser=p.type,
        output=p.output,
        nproc=p.processes,
        quiet=p.quiet,
        stream=p.stream,
        descriptor=p.descriptor,
        with3D=p.with3D,
        verbosity=p.verbosity,
    )


if __name__ == "__main__":
    freeze_support()
    main()



from collections import Iterable

import numpy as np

from mordred import Result, Descriptor, error, is_missing
from nose.tools import eq_, ok_, raises


class Dummy1(Descriptor):
    def __str__(self):
        return "Dummy1"

    def parameters(self):
        return ()

    def calculate(self):
        return 1


class Dummy2(Descriptor):
    def __str__(self):
        return "Dummy2"

    def parameters(self):
        return ()

    def calculate(self):
        raise ValueError("error")


result = Result(None, [1, error.Error(ValueError("error"), [])], [Dummy1(), Dummy2()])


def test_length():
    eq_(len(result), 2)


def test_fill_missing():
    assert np.isnan(result.fill_missing()[1])


def test_drop_missing():
    eq_(len(result.drop_missing()), 1)


def test_items():
    i = result.items()
    yield ok_, isinstance(i, Iterable)
    li = list(i)
    yield eq_, len(li), 2
    yield eq_, len(li[0]), 2

    d, v = li[0]
    yield ok_, isinstance(d, Descriptor)
    yield ok_, isinstance(v, int)


def test_keys():
    i = result.keys()
    yield ok_, isinstance(i, Iterable)
    li = list(i)
    yield eq_, len(li), 2

    yield ok_, isinstance(li[0], Descriptor)


def test_iter():
    i = iter(result)
    yield ok_, isinstance(i, Iterable)
    li = list(i)
    yield eq_, len(li), 2

    yield ok_, isinstance(li[0], int)


def test_reversed():
    i = reversed(result)
    yield ok_, isinstance(i, Iterable)
    li = list(i)
    yield eq_, len(li), 2

    yield ok_, isinstance(li[1], int)


def test_asdict_non_rawkey():
    d = result.asdict()
    yield eq_, len(d), 2
    yield ok_, isinstance(d, dict)
    for k in d.keys():
        yield ok_, isinstance(k, str)


def test_asdict_rawkey():
    d = result.asdict(True)
    yield eq_, len(d), 2
    yield ok_, isinstance(d, dict)
    for k in d.keys():
        yield ok_, isinstance(k, Descriptor)


def test_ix():
    yield eq_, 1, result.ix[0]
    yield ok_, is_missing(result.ix[1])


def test_name():
    yield eq_, 1, result.name["Dummy1"]
    yield ok_, is_missing(result.name["Dummy2"])


def test_getitem():
    yield eq_, 1, result[0]
    yield ok_, is_missing(result[1])
    yield eq_, 1, result["Dummy1"]
    yield ok_, is_missing(result["Dummy2"])


@raises(TypeError)
def test_getitem_raise():
    result[1.2]



import io
import os
import sys

from setuptools import setup, find_packages

install_requires = ["six==1.*", "numpy==1.*", "networkx==2.*"]

if sys.version_info < (3, 4, 0):
    install_requires.append("enum34")


def get_version():
    with open(os.path.join("mordred", "_version.txt")) as f:
        return f.read().strip()


def get_test_data():
    for p, _, fs in os.walk(os.path.join("mordred", "tests", "references")):
        p = p.split(os.sep)[2:]

        for f in fs:
            yield os.path.join(*(p + [f]))


README_rst = ""
fndoc = os.path.join(os.path.dirname(__file__), "README.rst")
with io.open(fndoc, mode="r", encoding="utf-8") as fd:
    README_rst = fd.read()

setup(
    name="mordred",
    version=get_version(),
    description="molecular descriptor calculator",
    long_description=README_rst,
    license="BSD-3-Clause",
    author="Hirotomo Moriwaki",
    author_email="philopon.dependence@gmail.com",
    url="https://github.com/mordred-descriptor/mordred",
    platforms=["any"],
    keywords="QSAR chemoinformatics",
    packages=find_packages(),
    package_data={
        "mordred": ["data/*.txt", "_version.txt"],
        "mordred.tests": list(get_test_data()),
    },
    install_requires=install_requires,
    tests_require=["nose==1.*", "PyYaml>=4.2b1"],
    extras_require={"full": ["pandas", "tqdm"]},
    cmdclass={"test": None},
)



import rdkit.Chem.GraphDescriptors as RDKit

from ._base import Descriptor
from ._graph_matrix import DistanceMatrix

__all__ = ("BertzCT",)


class BertzCT(Descriptor):
    r"""Bertz CT descriptor(rdkit wrapper)."""

    since = "1.0.0"
    __slots__ = ()
    explicit_hydrogens = False

    def description(self):
        return "Bertz CT"

    @classmethod
    def preset(cls, version):
        yield cls()

    def __str__(self):
        return self.__class__.__name__

    def parameters(self):
        return ()

    def dependencies(self):
        return {"D": DistanceMatrix(self.explicit_hydrogens)}

    def calculate(self, D):
        return float(RDKit.BertzCT(self.mol, dMat=D))

    rtype = float



from ._base import Descriptor

__all__ = ("AromaticAtomsCount", "AromaticBondsCount")


class AromaticBase(Descriptor):
    __slots__ = ()

    @classmethod
    def preset(cls, version):
        yield cls()

    def __str__(self):
        return self._name

    def parameters(self):
        return ()

    rtype = int


class AromaticAtomsCount(AromaticBase):
    r"""aromatic atoms count descriptor."""

    since = "1.0.0"
    __slots__ = ()
    _name = "nAromAtom"

    def description(self):
        return "aromatic atoms count"

    def calculate(self):
        return sum(1 for a in self.mol.GetAtoms() if a.GetIsAromatic())


class AromaticBondsCount(AromaticBase):
    r"""aromatic bonds count descriptor."""

    since = "1.0.0"
    __slots__ = ()
    _name = "nAromBond"

    def description(self):
        return "aromatic bonds count"

    def calculate(self):
        return sum(1 for b in self.mol.GetBonds() if b.GetIsAromatic())



import numpy as np

from nose.tools import eq_, ok_
from mordred.error import Error, Missing
from mordred._base.pandas_module import MordredDataFrame

df = MordredDataFrame(
    [
        [1, 2],
        [1, Error(ValueError("error"), ())],
        [Missing(ValueError("missing"), ()), 4],
    ]
)


def test_copy():
    ok_(isinstance(df.copy(), MordredDataFrame))


def test_fill_missing():
    f = df.fill_missing()
    yield eq_, f.ix[0, 0], 1
    yield eq_, f.ix[0, 1], 2
    yield eq_, f.ix[1, 0], 1
    yield ok_, np.isnan(f.ix[1, 1])
    yield ok_, np.isnan(f.ix[2, 0])
    yield eq_, f.ix[2, 1], 4



import numpy as np
from rdkit import Chem

from ._base import Descriptor


class DistanceMatrix(Descriptor):
    __slots__ = ("explicit_hydrogens", "useBO", "useAtomWts")

    hermitian = True

    def parameters(self):
        return self.explicit_hydrogens, self.useBO, self.useAtomWts

    def __init__(self, explicit_hydrogens, useBO=False, useAtomWts=False):
        self.explicit_hydrogens = explicit_hydrogens
        self.useBO = useBO
        self.useAtomWts = useAtomWts

    def calculate(self):
        return Chem.GetDistanceMatrix(
            self.mol, useBO=self.useBO, useAtomWts=self.useAtomWts, force=True
        )


class Eccentricity(DistanceMatrix):
    __slots__ = ()

    def dependencies(self):
        return {
            "D": DistanceMatrix(self.explicit_hydrogens, self.useBO, self.useAtomWts)
        }

    def calculate(self, D):
        return D.max(axis=0)


class Radius(Eccentricity):
    __slots__ = ()

    def dependencies(self):
        return {"E": Eccentricity(self.explicit_hydrogens, self.useBO, self.useAtomWts)}

    def calculate(self, E):
        return E.min()


class Diameter(Eccentricity):
    __slots__ = ()

    def calculate(self, D):
        return D.max()


class AdjacencyMatrix(Descriptor):
    __slots__ = ("explicit_hydrogens", "useBO", "order")

    hermitian = True

    def parameters(self):
        return self.explicit_hydrogens, self.useBO, self.order

    def __init__(self, explicit_hydrogens, useBO=False, order=1):
        self.explicit_hydrogens = explicit_hydrogens
        self.useBO = useBO
        self.order = order

    def dependencies(self):
        if self.order > 1:
            return {
                "A1": self.__class__(self.explicit_hydrogens, self.useBO, 1),
                "An": self.__class__(
                    self.explicit_hydrogens, self.useBO, self.order - 1
                ),
            }
        else:
            return {}

    def calculate(self, An=None, A1=None):
        if self.order == 1:
            return Chem.GetAdjacencyMatrix(self.mol, useBO=self.useBO, force=True)

        return An.dot(A1)


class Valence(AdjacencyMatrix):
    __slots__ = ()

    def dependencies(self):
        return {"D": AdjacencyMatrix(self.explicit_hydrogens, self.useBO)}

    def calculate(self, D):
        return D.sum(axis=0)


class DistanceMatrix3D(Descriptor):
    __slots__ = ("explicit_hydrogens", "useAtomWts")

    require_3D = True

    def parameters(self):
        return self.explicit_hydrogens, self.useAtomWts

    def __init__(self, explicit_hydrogens, useAtomWts=False):
        self.explicit_hydrogens = explicit_hydrogens
        self.useAtomWts = useAtomWts

    def calculate(self):
        return np.sqrt(np.sum((self.coord[:, np.newaxis] - self.coord) ** 2, axis=2))


class Eccentricity3D(DistanceMatrix3D):
    __slots__ = ()

    def dependencies(self):
        return {"D": DistanceMatrix3D(self.explicit_hydrogens, self.useAtomWts)}

    def calculate(self, D):
        return D.max(axis=0)


class Radius3D(Eccentricity3D):
    __slots__ = ()

    def dependencies(self):
        return {"E": Eccentricity3D(self.explicit_hydrogens, self.useAtomWts)}

    def calculate(self, E):
        return E.min()


class Diameter3D(Eccentricity3D):
    __slots__ = ()

    def calculate(self, D):
        return D.max()



import json

from mordred import Calculator, descriptors
from nose.tools import eq_


def test_json():
    calc = Calculator(descriptors)
    j = json.dumps(calc.to_json())
    calc2 = Calculator.from_json(json.loads(j))
    eq_(calc.descriptors, calc2.descriptors)



from ._base import Descriptor
from ._graph_matrix import Valence

__all__ = ("ZagrebIndex",)


class ZagrebIndex(Descriptor):
    r"""Zagreb index descriptor.

    .. math::

        {}^\lambda M_1 = \sum_{atoms} d_i^\lambda

        {}^\lambda M_2 = \sum_{edges} \left(d_i \cdot d_j \right)^\lambda

    where
    :math:`d_i` is degree of i-th atom

    :type version: int
    :param version: Zagreb index version. 1 or 2.

    :type variable: int
    :param variable: lambda value.

    :returns: NaN when valence of any atoms are 0
    """

    since = "1.0.0"
    __slots__ = ("_version", "_variable")
    explicit_hydrogens = False

    def description(self):
        if self._variable == 1:
            return "Zagreb index (version {})".format(self._version)
        elif self._variable == -1:
            return "modified Zagreb index (version {})".format(self._version)
        else:
            return "Zagreb like index (lambda = {}, version {})".format(
                self._variable, self._version
            )

    @classmethod
    def preset(cls, version):
        return (cls(v, x) for x in [1, -1] for v in [1, 2])

    def __str__(self):
        if self._variable in {1, -1}:
            m = "" if self._variable == 1 else "m"
            return "{}Zagreb{}".format(m, self._version)

        return "Zagreb{}_{}".format(self._version, self._variable)

    def parameters(self):
        return self._version, self._variable

    def __init__(self, version=1, variable=1):
        assert version in {1, 2}
        self._version = version
        self._variable = variable

    def dependencies(self):
        return {"V": Valence(self.explicit_hydrogens)}

    def calculate(self, V):
        V = V.astype("float")

        if self._version == 1:
            with self.rethrow_zerodiv():
                return (V ** (self._variable * 2)).sum()
        else:
            return float(
                sum(
                    (V[b.GetBeginAtomIdx()] * V[b.GetEndAtomIdx()]) ** self._variable
                    for b in self.mol.GetBonds()
                )
            )

    rtype = float



"""All descriptor modules are loaded.

.. code:: python

    >>> from mordred import Calculator, descriptors

    >>> descriptors.ABCIndex.__name__  # ABCIndex module
    'mordred.ABCIndex'

    >>> len(descriptors.all)  # all descriptor modules
    50

    >>> calc = Calculator(descriptors) # all descriptors
"""


def _import_all_descriptors():
    import os
    from importlib import import_module
    from .._base.descriptor import is_descriptor_class

    names = []
    values = []
    base_dir = os.path.dirname(os.path.dirname(__file__))

    for name in sorted(os.listdir(base_dir)):
        name, ext = os.path.splitext(name)
        if name[:1] == "_" or ext != ".py":
            continue

        mdl = import_module(".." + name, __package__)

        if any(v for v in mdl.__dict__.values() if is_descriptor_class(v)):
            names.append(name)
            values.append(mdl)
            globals()[name] = mdl

        globals()["__all__"] = tuple(names)
        globals()["all"] = tuple(values)


_import_all_descriptors()



from mordred import Descriptor


class Dummy1(Descriptor):
    pass


class Dummy2(Dummy1):
    since = "1.0.0"

    def parameters(self):
        return ()

    def calculate(self):
        return 0.0

    rtype = float


class Dummy3(Dummy2):
    @classmethod
    def preset(cls, version):
        yield cls()


class Dummy4(Dummy3):
    def __init__(self, i=0):
        self.i = i

    def __eq__(self, other):
        return isinstance(other, Dummy4) and self.i == other.i

    def __str__(self):
        return "Dummy4_{}".format(self.i)

    @classmethod
    def preset(cls, version):
        yield cls(0)
        yield cls(1)
